/* hanabi.c
 * Copyright (c) 1995-2004 Julio M. Otuyama - otuyama@inf.ufsc.br
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdlib.h>
#include <math.h>
#include "screenhack.h"

/********** XWindows **********/
static struct xw {
  Display *xdisplay;
  Window xwindow;
  GC xgc;
  Colormap colormap;
  int width;
  int height;
  XColor black;
} xw;

void xw_static_init (Display *xdisplay, Window xwindow) {
  GC xgc;

  XWindowAttributes xgwa;
  XGetWindowAttributes (xdisplay, xwindow, &xgwa);
  xw.colormap = xgwa.colormap;
  xw.width    = xgwa.width;
  xw.height   = xgwa.height;

  xgc= XCreateGC (xdisplay, xwindow, 0, 0);

  xw.xdisplay = xdisplay;
  xw.xwindow = xwindow;
  xw.xgc = xgc;

  xw.black.red = xw.black.green = xw.black.blue = 50;
  xw.black.flags = DoRed|DoGreen|DoBlue;

  XClearWindow (xw.xdisplay, xw.xwindow); 
}

void setColor(XColor color) {
  if (XAllocColor (xw.xdisplay,xw.colormap, &color)) {
    XSetForeground (xw.xdisplay, xw.xgc, color.pixel);
  }
}

#define drawPoint(x,y) (XDrawRectangle(xw.xdisplay, xw.xwindow, xw.xgc, x, y, 1,1))

/********** Utilities **********/
#define max(x,y) ((x)>(y)? (x):(y))
#define min(x,y) ((x)<(y)? (x):(y))
#define range(x,limmin,limmax) (max(min((x),(limmax)),(limmin)))

/********** Hanabi **********/
typedef int bool;

double normRandom(void) {
  return ((double)random()/RAND_MAX);
}

/******************** Explosion state ********************/
typedef struct State {
  long sleep;
  long launch;
  long explode;
  long disappear;
	
  long counter;
}State;

/* state init, change */
void State_init(State*s,long wait) {
  s->sleep     = (long)(3*normRandom()+20*wait);
  s->launch    = 3;
  s->explode   = 90;
  s->disappear = (long)(2+5*normRandom());

  s->counter = 0;
}

void State_update(State*s) { s->counter += 1; }

/* State */
bool State_sleeping(State*s)     { return ((0<=s->counter)&&(s->counter<s->sleep));                                                                   }
bool State_launching(State*s)    { return ((0+s->sleep<=s->counter)&&(s->counter<s->sleep+s->launch));                                                }
bool State_exploding(State*s)    { return ((0+s->sleep+s->launch<=s->counter)&&(s->counter<s->sleep+s->launch+s->explode));                           }
bool State_disappearing(State*s) { return ((0+s->sleep+s->launch+s->disappear<=s->counter)&&(s->counter<s->sleep+s->launch+s->disappear+s->explode)); }
bool State_finish(State*s)       { return (0+s->sleep+s->launch+s->disappear+s->explode<=s->counter);                                                 }

/* Explosion attribute */
double State_radius_explode(State*s)   { return (((double)s->counter-(s->sleep+s->launch))/(s->explode));              }
double State_radius_disappear(State*s) { return (((double)s->counter-(s->sleep+s->launch+s->disappear))/(s->explode)); }

/******************** Explosion ********************/
static int Explosion_num_fragment=150;
static double Explosion_fragment[150][3];
static double Explosion_st_cycle;

typedef struct Explosion {
  double r_size;
  double pos_x;
  double pos_y;
  double pos_z;
  XColor color1;
  XColor color2;
  State state;
} Explosion;

void Explosion_static_init(int st_cycle) {
  int i;

  Explosion_st_cycle = st_cycle;

  for (i=0; i<Explosion_num_fragment; i++) {
    double x,y,z,r;
    do {
      x=2*normRandom()-1;
      y=2*normRandom()-1;
      z=2*normRandom()-1;
      r=sqrt(x*x+y*y+z*z);
    } while (!((0.6<=r)&&(r<=0.8))); /* uniform distribution on sphere */
    Explosion_fragment[i][0]=x/r;
    Explosion_fragment[i][1]=y/r;
    Explosion_fragment[i][2]=z/r;
  }
	
}

void Explosion_init(Explosion*explos,long wait) {
  explos->r_size = 0.9+0.6*normRandom();
  explos->pos_x = normRandom();
  explos->pos_y = normRandom();
  explos->pos_z = 2*normRandom();
  hsv_to_rgb (random () % 360, 1.0, 1.0, &explos->color1.red, &explos->color1.green, &explos->color1.blue);
  explos->color1.flags = DoRed | DoGreen | DoBlue;
  hsv_to_rgb (random () % 360, 1.0, 1.0, &explos->color2.red, &explos->color2.green, &explos->color2.blue);
  explos->color2.flags = DoRed | DoGreen | DoBlue;
  State_init(&explos->state,wait);
}

void Explosion_plot (Explosion*explos, double x, double y, double z) { /* stereo features */
  int i;
  double stx,sty;

  for (i=-1; i<=Explosion_st_cycle; i++) {
    stx = (
      (
        x*explos->r_size                 /* horizontal explosion radius */
        *(1/(2*Explosion_st_cycle))      /* stereo constant (for repetition) */
      )+(
        i*(1/(Explosion_st_cycle))       /* stereo repetition */
      )+(
        0.005                            /* stereo adjust */
        *(
          z*explos->r_size               /* disparity: explosion radius - deep */
          +4*(explos->pos_z-0.5)         /* disparity: random deep */
        )
        *(i-(Explosion_st_cycle+1)/2)    /* disparity factor (repetition) */
      )+(
        explos->pos_x*(1/(2*Explosion_st_cycle)) /* random horizontal position */
      )
    );
    sty = (
      (
        y*explos->r_size                 /* vertical explosion radius */
        *(1/(2*Explosion_st_cycle))*xw.width/xw.height /* aspect ratio */
      )+(
        explos->pos_y                    /* random vertical position */
      )
    );
    drawPoint((int)(xw.width*stx),(int)(xw.height*sty)); 
  }
}

void Explosion_draw (Explosion*explos, XColor c1, XColor c2, double r) { /* other features */
  int n;
  int i;
  double radius = 1 - exp(-r*5); /* explosion speed effect */
  double limiar = 0.8;
  double gravity;
  if (radius<limiar) {
    n = Explosion_num_fragment;
  } else {
    n = (int)(Explosion_num_fragment*(1-radius)/(1-limiar)); /* blink effect */
  }
	 
  gravity = 0.5*radius*radius; /* gravity effect */
  setColor(c1);
  for (i=0; i<n; i+=2) {
    Explosion_plot(
      explos,
      Explosion_fragment[i][0]*radius,
      Explosion_fragment[i][1]*radius+gravity,
      Explosion_fragment[i][2]*radius
    );
  }
  setColor(c2);
  for (i=1; i<n; i+=2) {
    Explosion_plot(
      explos,
      Explosion_fragment[i][0]*radius,
      Explosion_fragment[i][1]*radius+gravity,
      Explosion_fragment[i][2]*radius
    );
  }
}

void Explosion_update(Explosion*explos) {
  if (State_exploding(&explos->state)) {
    Explosion_draw(explos,explos->color1,explos->color2,State_radius_explode(&explos->state)); 
  }
  if (State_disappearing(&explos->state)) {
    Explosion_draw(explos,xw.black,xw.black,State_radius_disappear(&explos->state)); 
  }
  State_update(&explos->state);
}

bool Explosion_finish(Explosion*explos) {
  return State_finish(&explos->state);
}

/******************** Main ********************/
static struct Hanabi {
  int num_explos;
  Explosion vExplosion[10];
} Hanabi;

void Hanabi_init(void) {
  int i;
  int num_explos; 
	
  num_explos = get_integer_resource ("num_explosion", "Integer");
  Hanabi.num_explos = range(num_explos,1,10);;
  for (i=0; i<Hanabi.num_explos; i++) {
    Explosion_init(&(Hanabi.vExplosion[i]),i);
  }

}

void Hanabi_update(void) {
  int i;

  for (i=0; i<Hanabi.num_explos; i++) {
    Explosion_update(&(Hanabi.vExplosion[i]));
    if (Explosion_finish(&(Hanabi.vExplosion[i]))) {
      Explosion_init(&(Hanabi.vExplosion[i]),0);
    }
  }
}

/********** Screenhack **********/
char *progclass = "Hanabiprogclass";

char *defaults [] = {
  ".background:	black",
  "*fps:           40",
  "*num_explosion: 6",
  "*st_cycle:      5",
  0
};

XrmOptionDescRec options [] = {
  { "-fps",           ".fps",	        XrmoptionSepArg, 0 },
  { "-num_explosion", ".num_explosion", XrmoptionSepArg, 0 },
  { "-st_cycle",      ".st_cycle",      XrmoptionSepArg, 0 },
  { 0, 0, 0, 0 }
};

void screenhack (Display *xdisplay, Window xwindow) {
  int delay;
  int st_cycle;
  int fps;

  xw_static_init (xdisplay, xwindow);
  st_cycle = range (get_integer_resource ("st_cycle", "Integer"),3,8);
  Explosion_static_init(st_cycle);
  fps = range(get_integer_resource ("fps", "Integer"),5,60);
  delay = 1000000/fps;

  Hanabi_init ();
  while (1) {
    Hanabi_update ();

    screenhack_handle_events (xdisplay);
    usleep (delay);
  }

}
